(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2005-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: layout.ml,v 1.14 2017/05/31 15:54:22 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module L1 = Ulist1;;
module A = Uarray;;

let _0_ = Nat._0_;;

type t =
   | Permutation of Nat.t L1.t
   | C
   | F
;;

let layout_to_ob ob = function
  | C -> Buffer.add_string ob "C"
  | F -> Buffer.add_string ob "F"
  | Permutation (i, is) ->
    let item_to_ob ob i =
      Buffer.add_string ob (Nat.to_string i);
      Buffer.add_string ob ", " in
    L.iter (item_to_ob ob) (i :: is)
;;

let permutation_of_ib ib =
  let item_of_ib ib = Scanf.bscanf ib "%d" in
  let rec items_of_ib ib accu =
    if Scanf.Scanning.end_of_input ib then L.rev accu else
    Scanf.bscanf ib "%0c" (function
    | '0' .. '9' ->
      item_of_ib ib (fun it -> items_of_ib ib (Nat.unsafe_of_int it :: accu))
    | ',' ->
      Scanf.bscanf ib ",\032" ();
      items_of_ib ib accu
    | _ -> L.rev accu) in
  Scanf.bscanf ib "%0c" (function
  | '0' .. '9' ->
    item_of_ib ib
      (fun it -> Permutation (Nat.unsafe_of_int it, items_of_ib ib []))
  | c ->
    failwith
      (Printf.sprintf "permutation_of_ib: unexpected char %c" c))
;;

let layout_of_ib ib =
  Scanf.bscanf ib "%0c" (function
  | 'C' -> Scanf.bscanf ib "%_c" C
  | 'F' -> Scanf.bscanf ib "%_c" F
  | '0' .. '9' -> permutation_of_ib ib
  | c -> failwith (Printf.sprintf "layout_of_ib: unexpected char %c" c))
;;

let check_layout p l =
  if p = _0_ then
    failwith "check_layout: dimension should be positive." else
  match l with
  | C | F -> ()
  | Permutation (i, []) ->
    if i > _0_ then
      failwith
        (Printf.sprintf
           "check_layout: singleton permutation should be [0] not [%d]."
           (Nat.to_int i))
  | Permutation (i, l) ->
    let lperm = Nat.succ (L.length l) in
    if lperm <> p then
      failwith
        (Printf.sprintf
           "check_layout: permutation size %d \
            and array dimension %d mismatch."
           (Nat.to_int lperm) (Nat.to_int p)) else
    let v = A.make p false in
    let check_item i =
      if i >= p then
        failwith
          (Printf.sprintf
             "check_layout: permutation item %d is out of range."
             (Nat.to_int i)) else
      if A.unsafe_get v i then
        failwith
          (Printf.sprintf
             "check_layout: permutation item %d found twice."
             (Nat.to_int i)) in
    L.iter (fun i -> check_item i; A.unsafe_set v i true) (i :: l);;

let check = check_layout
and to_ob = layout_to_ob
and of_ib = layout_of_ib
;;
