(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2005-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: pio.ml,v 1.45 2017/05/31 15:54:22 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module L1 = Ulist1;;
module S = Ustring;;
module A = Uarray;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module D = Udebug;;
module U = Uunix;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

module Tv = Typed_value;;
module Wtv = Write_typed_value;;
module Com = Communication;;
module Wc = Write_communication;;
module Rc = Read_communication;;

(* General sending communication functions to a [Buffer.t] buffer. *)

let send_communication = Wc.write_communication;;

let send_values ob vals = send_communication ob (Com.Pio_Phrase vals);;

let send_one_value ob v = send_values ob [| v |];;

(* Type specialized sending communication functions to a [Buffer.t] buffer. *)

let gen_send of_native ob t = Wtv.write_typed_value ob (of_native t);;

let send_string = gen_send Tv.of_string
and send_int = gen_send Tv.of_int
and send_int_couple = gen_send Tv.of_int_couple
and send_int_triple = gen_send Tv.of_int_triple
and send_float = gen_send Tv.of_float
and send_float_couple = gen_send Tv.of_float_couple
and send_float_triple = gen_send Tv.of_float_triple
and send_string_vector = gen_send Tv.of_string_vector
and send_int_vector = gen_send Tv.of_int_vector
and send_int_couple_vector = gen_send Tv.of_int_couple_vector
and send_int_triple_vector = gen_send Tv.of_int_triple_vector
and send_float_vector = gen_send Tv.of_float_vector
and send_float_couple_vector = gen_send Tv.of_float_couple_vector
and send_float_triple_vector = gen_send Tv.of_float_triple_vector
and send_c_int_matrix = gen_send Tv.of_c_int_matrix
and send_fortran_int_matrix = gen_send Tv.of_fortran_int_matrix
and send_c_float_matrix = gen_send Tv.of_c_float_matrix
and send_fortran_float_matrix = gen_send Tv.of_fortran_float_matrix
;;

(* General receiving communication functions from a
   [Scanf.Scanning.in_channel] scanning buffer. *)

exception Pio_closed_input_channel = Rc.Pio_closed_input_channel;;

let receive_communication = Rc.read_communication;;

let receive_values ib =
  match receive_communication ib with
  | Com.Pio_Phrase vals -> vals
  | Com.Pio_Task _ | Com.Pio_Result _ | Com.Pio_Error _ | Com.Pio_Service _ ->
    failwith "Pio.receive_values: communication is not a phrase"
;;

let receive_one_value ib =
  let vals = receive_values ib in
  if A.length vals = _1_ then A.unsafe_get vals _0_ else
  failwith
    "Pio.receive_one_value: \
     communication phrase has more than one value"
;;

(* Type specialized receiving communication functions. *)

let gen_receive to_native ib =
  try to_native (Read_typed_value.read_typed_value ib) with
  | End_of_file ->
    prerr_endline "Pio: premature end of communication";
    exit 2
  | x ->
    prerr_endline
      (Printf.sprintf "Pio: spurious exception: %s" (Printexc.to_string x));
    exit 2
;;

let receive_string = gen_receive Tv.to_string
and receive_int = gen_receive Tv.to_int
and receive_int_couple = gen_receive Tv.to_int_couple
and receive_int_triple = gen_receive Tv.to_int_triple
and receive_float = gen_receive Tv.to_float
and receive_float_couple = gen_receive Tv.to_float_couple
and receive_float_triple = gen_receive Tv.to_float_triple
and receive_string_vector = gen_receive Tv.to_string_vector
and receive_int_vector = gen_receive Tv.to_int_vector
and receive_int_couple_vector = gen_receive Tv.to_int_couple_vector
and receive_int_triple_vector = gen_receive Tv.to_int_triple_vector
and receive_float_vector = gen_receive Tv.to_float_vector
and receive_float_couple_vector = gen_receive Tv.to_float_couple_vector
and receive_float_triple_vector = gen_receive Tv.to_float_triple_vector
and receive_c_int_matrix = gen_receive Tv.to_c_int_matrix
and receive_fortran_int_matrix = gen_receive Tv.to_fortran_int_matrix
and receive_c_float_matrix = gen_receive Tv.to_c_float_matrix
and receive_fortran_float_matrix =
  gen_receive Tv.to_fortran_float_matrix
;;

(* Foreman/Worker. *)

type pf = D.printing_facilities;;
type spf = D.specialized_printing_facilities;;

type values = Com.values
and service_name = Com.service_name
and task_name = Com.task_name
and communication = Com.t
;;

type 'a injector = 'a -> values
and 'a projector = values -> 'a
;;

type ('a, 'b) task_wrapper =
  {
   fname : Perv.function_name;
   tname : task_name;
   spf : spf;
   inj : 'a injector;
   proj : 'b projector;
  }
;;

type ('a, 'b) task = spf -> 'a -> 'b;;

type welcome_action = Perv.gen_message
and farewell_action = Perv.gen_message
;;

type task_computation = pf -> U.process_io_buffers -> values -> unit;;

type task_table = (task_name * task_computation) L1.t;;

exception Unbound_communication of communication;;
exception Unbound_service_communication of (service_name * values);;
exception Unbound_task of task_name;;
exception Uninitialized_worker of task_name;;

let task_prefix = Printf.sprintf "Task %S:";;

let make_task_wrapper tname mk_inj mk_proj pf fname =
  let spf =
    D.prefix_messages (task_prefix tname) (D.set_function_name fname pf) in
  {
   fname = fname;
   tname = tname;
   spf = spf;
   inj = mk_inj spf;
   proj = mk_proj spf;
  }
;;

let gen_make_pf sname =
  let oks = Ok.make_option_keys sname Ok.Automatic in
  let skd = { Co.name = sname; option_key_prefixes_opt = Some oks; } in
  D.application_printing_facilities_with_static_tag skd
;;

let foreman_make_printing_facilities () = gen_make_pf "Foreman"
and worker_make_printing_facilities () = gen_make_pf "Worker"
;;

let worker_test_wrap_task piob tname inj input =
  let { U.output_buffer = ob; flush_output = flush; _; } = piob in
  let input_values = inj input in
  send_communication ob (Com.Pio_Task (tname, input_values));
  flush ();
;;

let foreman_wrap_task cmd piob_ref tw input =
  let {
        fname = fname; tname = tname;
        spf = { D.spf_report = report; _; } as spf;
        inj = inj;
        proj = proj;
      } = tw in
  let ({ U.input_buffer = ib; _; } as piob) = U.pio_spawn spf cmd piob_ref in
  report "started";
  worker_test_wrap_task piob tname inj input;
  report "task sent";
  let output =
    match receive_communication ib with
    | Com.Pio_Result output_values ->
        report "result received";
        proj output_values
    | Com.Pio_Error exns ->
        let sl = A.to_list (A.map Tv.to_text exns) in
        let msg = S.concat " " sl in
        Perv.failwith fname msg
    | Com.Pio_Service s -> raise (Unbound_service_communication s)
    | Com.Pio_Phrase _ | Com.Pio_Task _ as com ->
        raise (Unbound_communication com) in
  report "finished";
  output
;;

let worker_wrap_task piob tw task input_values =
  let {
        spf = { D.spf_report = report; _; } as spf;
        inj = inj;
        proj = proj; _;
      } = tw in
  let { U.output_buffer = ob; flush_output = flush; _; } = piob in
  report "started";
  let input = proj input_values in
  let output = task spf input in
  report "result computed";
  let output_values = inj output in
  send_communication ob (Com.Pio_Result output_values);
  flush ();
  report "result sent back";
  report "finished";
;;

let spread_out_task tname tbl pf piob vals =
  let compute =
    try L1.assoc tname tbl with
    | Not_found -> raise (Unbound_task tname) in
  compute pf piob vals
;;

let spread_out_communication tbl pf piob =
  let { D.report; _; } = pf in
  let { U.input_buffer = ib; _; } = piob in
  match receive_communication ib with
  | Com.Pio_Task (tname, vals) ->
      report (Perv.prefix (task_prefix tname) "received");
      spread_out_task tname tbl pf piob vals
  | Com.Pio_Service s -> raise (Unbound_service_communication s)
  | Com.Pio_Phrase _ | Com.Pio_Result _ | Com.Pio_Error _ as com ->
    raise (Unbound_communication com)
;;

let rec loop tbl pf piob =
  spread_out_communication tbl pf piob;
  loop tbl pf piob
;;

let init_required tname =
  Perv.prefix (task_prefix tname) "requires initialization"
;;

let spurious_exn exn =
  Printf.sprintf
    "Spurious exception: %S, please report"
    (Printexc.to_string exn)
;;

let worker_main hi_action bye_action tbl =
  let ({ D.report = report; error = error; _; } as pf) =
    worker_make_printing_facilities () in
  Co.parse ();
  let print_error_and_exit msg = error None msg; exit 1 in
  report (hi_action ());
  let piob = U.std_process_io_buffers in
  begin
    try loop tbl pf piob with
    | Unbound_task _
    | Unbound_communication _
    | Unbound_service_communication _ as exn ->
        print_error_and_exit (Printexc.to_string exn)
    | Com.Wrong_communication msg -> print_error_and_exit msg
    | Uninitialized_worker tname -> print_error_and_exit (init_required tname)
    | Pervasives.Exit -> ()
      (* Exit is not raised by any standard library function, but could be
         explicitely raised by some worker task function, and it means we
         should exit. *)
    | Pio_closed_input_channel -> ()
      (* Our input channel has been closed:
         no way to read something again, we also give up. *)
    | exn -> print_error_and_exit (spurious_exn exn)
  end;
  report (bye_action ());
;;
