(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: helper.ml,v 1.16 2017/01/05 16:44:43 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;
module A = Uarray;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let scalar_prod = A.fold_left2 (fun accu ai bi -> accu + ai * bi) 0;;

open Parser;;
open Parser_char;;

let parse_vec p =
  pspaced (pchar '[') >>
  (pspaced (sep_by1 (pspaced (pchar_ign ';')) (pspaced p))) >>= fun l ->
  pspaced (pchar ']') >> return (A.of_list l)
;;

let parse_vector = parse_vec pint;;
let parse_matrix = parse_vec parse_vector;;
let parse_fvector = parse_vec pfloat;;
let parse_tridiag =
  parse_fvector >>= fun d0 ->
  parse_fvector >>= fun d1 ->
  parse_fvector >>= fun d2 ->
  return (Tridiag.create d0 d1 d2)
;;

let slurp ic =
  let rec loop lines =
    match
      try Some (input_line ic) with
      | End_of_file -> None
    with
    | None -> S.concat "\n" (L.rev lines)
    | Some l -> loop (l :: lines) in
  loop []
;;

let sanitize matopt =
  match matopt with
  | None -> None
  | Some mat ->
    let len = A.length mat in
    if len = _0_ then None else
    (* Now, we have len > 0.
       Therefore, the next A.unsafe_get will be safe. *)
    let n0 = A.length (A.unsafe_get mat _0_) in
    let rec loop i l =
      if i >= len then true else
      (* Now, we have i < len.
         Therefore, the next A.unsafe_get will be safe. *)
      A.length (A.unsafe_get mat i) = l && loop (Nat.succ i) l in
    if not (loop _1_ n0) then None else
    matopt
;;

module IO = struct
  let input_vector ic = parse parse_vector (slurp ic)
  and input_fvector ic = parse parse_fvector (slurp ic)
  and input_matrix ic = sanitize (parse parse_matrix (slurp ic))
  and input_tridiag ic = parse parse_tridiag (slurp ic)
  ;;

  let output_vector oc vec =
    let len = A.length vec in
    let ppf = Format.formatter_of_out_channel oc in
    let rec loop ppf i =
      if i >= len then () else
      (* Now, we have i < len.
         Therefore, the next A.unsafe_get will be safe. *)
      let vec_i = A.unsafe_get vec i in
      let i = Nat.succ i in
      if i = len then Format.fprintf ppf "%d" vec_i else
      Format.fprintf ppf "%d;@ %a" vec_i loop i in
    Format.fprintf ppf "@[<hv 2>[@ %a@ ]@]@." loop _0_
  ;;

  let output_raw oc vecs =
    let output_vec v =
      A.iter (Printf.fprintf oc "%f\n") v in
    A.iter output_vec vecs;
    flush oc;
  ;;
end
;;

let fetch_std_data mat_file vec_file =
  let parse_error = Printf.eprintf "Parsing error while processing %S.\n" in
  let mat_chan = open_in mat_file
  and vec_chan = open_in vec_file in
  match (IO.input_matrix mat_chan, IO.input_vector vec_chan) with
  | (None, None) -> parse_error mat_file; parse_error vec_file; exit 2
  | (None, Some _) -> parse_error vec_file; exit 2
  | (Some _, None) -> parse_error mat_file; exit 2
  | (Some mat, Some vec) -> (mat, vec)
;;

let fetch_tri_data mat_file vec_file =
  let parse_error = Printf.eprintf "Parsing error while processing %S.\n" in
  let mat_chan = open_in mat_file
  and vec_chan = open_in vec_file in
  match (IO.input_tridiag mat_chan, IO.input_fvector vec_chan) with
  | (None, None) -> parse_error mat_file; parse_error vec_file; exit 2
  | (None, Some _) -> parse_error vec_file; exit 2
  | (Some _, None) -> parse_error mat_file; exit 2
  | (Some mat, Some vec) -> (mat, vec)
;;
