(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tridiag.ml,v 1.16 2017/01/04 11:58:31 fclement Exp $ *)

module Nat = Unatural;;
module A = Uarray;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
;;

type elmt = float
and top = elmt
and bot = elmt
and elmts = elmt array
and index = int
;;

type t = elmts * elmts * elmts
and tridiag = t
;;

type pvect =
  | Top of elmts * bot
  | Mid of index * top * elmts * bot
  | Bot of top * elmts
;;

let create diag0 diag1 diag2 =
  let n1 = A.length diag1 in
  if n1 = _0_ then invalid_arg "Tridiag.create" else
  (* Now, we have n1 > 0.
     Therefore, the next Nat.pred will not fail. *)
  if A.length diag0 <> A.length diag2 || A.length diag0 <> Nat.pred n1 then
    invalid_arg "Tridiag.create" else
  (diag0, diag1, diag2)
;;

let pmult (d0, d1, d2) pvect =
  let n = A.length d1 in
  let n = Nat.to_int n in
  let compute_middle f arr elmts start stop =
    for i = start to stop do
      arr.(i) <-
        elmts.(pred i) *. d0.(f (pred i)) +.
        elmts.(i) *. d1.(f i) +.
        elmts.(succ i) *. d2.(f i)
    done in
  match pvect with
  | Top (elmts, bot) ->
    let nelmts = A.length elmts in
    let out = A.make nelmts 0. in
    let nelmts = Nat.to_int nelmts in
    let pnelmts = pred nelmts in
    let ppnelmts = pred pnelmts in
    compute_middle (fun x -> x) out elmts 1 ppnelmts;
    out.(0) <- elmts.(0) *. d1.(0) +. elmts.(1) *. d2.(0);
    out.(pnelmts) <-
      elmts.(ppnelmts) *. d0.(ppnelmts) +.
      elmts.(pnelmts) *. d1.(pnelmts) +.
      bot *. d2.(pnelmts);
    out
  | Mid (idx, top, elmts, bot) ->
    let nelmts = A.length elmts in
    let out = A.make nelmts 0. in
    let nelmts = Nat.to_int nelmts in
    let pnelmts = pred nelmts in
    let ppnelmts = pred pnelmts in
    compute_middle (fun x -> idx + x) out elmts 1 (ppnelmts);
    out.(0) <-
      top *. d0.(pred idx) +.
      elmts.(0) *. d1.(idx) +.
      elmts.(1) *. d2.(idx);
    out.(pnelmts) <-
      elmts.(ppnelmts) *. d0.(idx + ppnelmts) +.
      elmts.(pnelmts) *. d1.(idx + pnelmts) +.
      bot *. d2.(idx + pnelmts);
    out
  | Bot (top, elmts) ->
    let nelmts = A.length elmts in
    let out = A.make nelmts 0. in
    let nelmts = Nat.to_int nelmts in
    let pnelmts = pred nelmts in
    let ppnelmts = pred pnelmts in
    compute_middle (fun x -> n - nelmts + x) out elmts 1 (ppnelmts);
    out.(0) <-
      top *. d0.(pred (n - nelmts)) +.
      elmts.(0) *. d1.(n - nelmts) +.
      elmts.(1) *. d2.(n - nelmts);
    out.(pnelmts) <-
      elmts.(ppnelmts) *. d0.(n - 2) +. elmts.(pnelmts) *. d1.(pred n);
    out
;;

let mk_pvect a =
  let chk b = if not b then invalid_arg "Tridiag.mk_pvect" in
  match Nat.to_int (A.length a) with
  | 0 -> invalid_arg "Tridiag.mk_pvect"
  | 1 ->
    let a = a.(0) in
    let l = A.length a in
    chk (l > _1_);
    (* Now, we have l >= 2.
       Therefore, the next Nat.sub will not fail. *)
    let elmts = A.sub a _1_ (Nat.sub l _2_)
    and bot = a.(pred (Nat.to_int l)) in
    [| Mid (1, a.(0), elmts, bot); |]
  | 2 ->
    let (a0, a1) = (a.(0), a.(1)) in
    let l0 = A.length a0
    and l1 = A.length a1 in
    chk (l0 > _0_ && l1 > _0_);
    [| Top (a0, a1.(0)); Bot (a0.(pred (Nat.to_int l0)), a1); |]
  | l ->
    let (a0, _al) = (a.(0), a.(pred l)) in
    let l2 = A.length a.(l - 2) in
    let out = A.make (Nat.of_int l) (Bot (0., [||])) in
    let idx = ref (Nat.to_int (A.length a0)) in
    for i = 1 to l - 2 do
      let li = A.length a.(pred i) in
      chk (li > _0_);
      out.(i) <-
        Mid (!idx, a.(pred i).(pred (Nat.to_int li)), a.(i), a.(succ i).(0));
      idx := !idx + Nat.to_int (A.length a.(i))
    done;
    out.(0) <- Top (a.(0), a.(1).(0));
    out.(pred l) <- Bot (a.(l - 2).(pred (Nat.to_int l2)), a.(pred l));
    out
;;

let split_vect n v =
  let n = Nat.of_int n in
  if n = _0_ then invalid_arg "Tridiag.split_vect" else
  let lv = A.length v in
  let l = Nat.div lv n in
  if l = _0_ then invalid_arg "Tridiag.split_vect" else
  let rec loop p i accu =
    (* Loop invariant: i = (n - p) * l.
       p is decreasing from n, and i is increasing from 0. *)
    if p <= _1_ then
      (* Now, we have p = 1 and i = (n - 1) * l < n * l <= lv.
         Therefore, the next Nat.sub will not fail. *)
      A.of_list (A.sub v i (Nat.sub lv i) :: accu) else
    (* Now, we have p > 1.
       Therefore, the next Nat.pred will not fail. *)
    loop (Nat.pred p) (Nat.add i l) (A.sub v i l :: accu) in
  loop n _0_ []
;;
