(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: IO_sockets.ml,v 1.38 2017/01/04 11:58:27 fclement Exp $ *)

module Nat = Unatural;;
module B = Ubytes;;

let _0_ = Nat._0_
and _16384_ = Nat.unsafe_of_int 16384
;;

(* In the following, integers output by Marshal and Unix functions are all
   nonnegative.
   Therefore, the next Nat.unsafe_of_int will al be safe. *)

let nat_marshal_header_size =
  Nat.unsafe_of_int Marshal.header_size

and nat_marshal_data_size buf ofs =
  Nat.unsafe_of_int (Marshal.data_size buf (Nat.to_int ofs))

and nat_marshal_from_string buf ofs =
  Marshal.from_string buf (Nat.to_int ofs)
;;

let nat_unix_read fd buf ofs len =
  Nat.unsafe_of_int (Unix.read fd buf (Nat.to_int ofs) (Nat.to_int len))

and nat_unix_write fd buf ofs len =
  Nat.unsafe_of_int (Unix.write fd buf (Nat.to_int ofs) (Nat.to_int len))

and nat_unix_openfile fname fl perms =
  Unix.openfile fname fl (Nat.to_int perms)

and nat_unix_mkfifo pname perms = Unix.mkfifo pname (Nat.to_int perms)
;;

let nat_unix_fork () =
  let pid = Unix.fork () in
  assert (pid >= 0);
  (* Now, we have pid >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  Nat.unsafe_of_int pid
;;

let bts = B.to_string;;
let bos = B.of_string;;

open Tools_common_types;;

(* Settings *)

let io_buff_default_size = ref _16384_;;

(* Types & exceptions *)

type 'a bufferized =
  {
   mutable buff : B.t;
   mutable cur : index;
   mutable len : length;
   chan : 'a;
  }
;;

type socket =
  {
   fd : Unix.file_descr;
   addr : Unix.sockaddr;
  }
;;

type fd =
  | Fd of Unix.file_descr
  | Socket of socket
;;

type ic = fd bufferized;;
type oc = fd;;

type t = ic * oc
and io = t
;;

exception IOError of string;;

(* Helpers *)

(* Execute procedure f until f is not interrupted during execution. *)
let rec intr_loop f =
  try f () with
  | Unix.Unix_error (Unix.EINTR, _, _) -> intr_loop f
;;

let ioerr msg = raise (IOError msg);;

let dump_sockaddr () = function
  | Unix.ADDR_UNIX addr -> addr
  | Unix.ADDR_INET (addr, port) ->
    Printf.sprintf "%s:%d" (Unix.string_of_inet_addr addr) port
;;

let print_sockaddr oc addr = Printf.fprintf oc "%s" (dump_sockaddr () addr);;

let read_sockaddr str =
  let sb = Scanf.Scanning.from_string str in
  let (addr, port) =
    try
      Scanf.bscanf sb "%s@:%d" (fun a p -> (Unix.inet_addr_of_string a, p))
    with
    | Scanf.Scan_failure _
    | Failure _
    | End_of_file ->
      failwith "Invalid address format, use www.xxx.yyy.zzz:port format" in
  Scanf.Scanning.close_in sb;
  Unix.ADDR_INET (addr, port)
;;

(* Bufferized data manipulation *)

let bufferize chan =
  {
   buff = B.create !io_buff_default_size;
   cur  = _0_;
   len  = _0_;
   chan = chan;
  }
;;

let realloc_string s sz =
  let len = B.length s in
  let s' = B.create sz in
  B.blit s _0_ s' _0_ (min len sz);
  s'
;;

let buff_consume_data buf feeder size =
  let { cur = c; len = l; _ } = buf in
  let sz =
    let oldsz = B.length buf.buff in
    if size <= oldsz then oldsz else begin
      buf.buff <- realloc_string buf.buff size;
      size
    end in
  if size <= l then
    (* Now, we have size <= l.
       Therefore, the next Nat.sub will not fail. *)
    let output = B.sub buf.buff c size in
    buf.cur <- Nat.add c size;
    buf.len <- Nat.sub l size;
    output
  else
    (* Now, we have l < size.
       Therefore, the next Nat.sub will not fail. *)
    let dsz = Nat.sub size l in
    if Nat.add size c < sz then (* <= ? *)
      let output = B.create size in
      let n = feeder buf dsz in
      B.blit buf.buff c output _0_ size;
      buf.cur <- Nat.add c size;
      assert (size <= Nat.add l n);
      (* Now, we have size <= l + n.
         Therefore, the next Nat.sub will not fail. *)
      buf.len <- Nat.sub (Nat.add l n) size;
      output
    else
      let output = B.create size in
      B.blit buf.buff c buf.buff _0_ l;
      buf.cur <- _0_;
      let n = feeder buf dsz in
      B.blit buf.buff _0_ output _0_ size;
      buf.cur <- size;
      assert (size <= Nat.add l n);
      (* Now, we have size <= l + n.
         Therefore, the next Nat.sub will not fail. *)
      buf.len <- Nat.sub (Nat.add l n) size;
      output
;;

(* Channel manipulations *)

let get_descr = function
  | Fd descr -> descr
  | Socket { fd = descr; _ } -> descr
;;

let sock_feeder buf min_sz =
  let buf_sz = B.length buf.buff in
  assert (Nat.add buf.cur buf.len <= buf_sz);
  (* Now, we have buf.cur + buf.len <= buf_sz.
     Therefore, the next Nat.sub will not fail. *)
  let max_sz = Nat.sub buf_sz (Nat.add buf.cur buf.len) in
  assert (max_sz >= min_sz);
  (* Now, we have max_sz >= min_sz.
     Therefore, the next Nat.sub will not fail. *)
  let descr = get_descr buf.chan in
  try
    let rec loop remaining =
      if remaining <= Nat.sub max_sz min_sz then
        (* Now, we have remainning <= max_sz - min_sz <= max_sz.
           Therefore, the next Nat.sub will not fail. *)
        Nat.sub max_sz remaining else
      let ofs = Nat.add buf.cur buf.len in
      let nread =
        intr_loop (fun () -> nat_unix_read descr buf.buff ofs remaining) in
      assert (nread <= remaining);
      (* Now, we have nread <= remaining.
         Therefore, the next Nat.sub will not fail. *)
      if nread = _0_ then raise End_of_file;
      buf.len <- Nat.add buf.len nread;
      loop (Nat.sub remaining nread) in
    loop max_sz
  with
  | Unix.Unix_error (e, _, _) ->
    ioerr ("IO.sock_feeder: " ^ Unix.error_message e)
;;

let send_to_channel ch str ofs n =
  let descr = get_descr ch in
  try
    let rec loop ofs len =
      let written = nat_unix_write descr str ofs len in
      if written < len then
        (* Now, we have written < len.
           Therefore, the next Nat.sub will not fail. *)
        loop (Nat.add ofs written) (Nat.sub len written) in
    loop ofs n
  with
  | Unix.Unix_error (e, _, _) ->
    ioerr ("IO.send_to_channel: " ^ Unix.error_message e)
;;

let send_bytes ch str = send_to_channel ch str _0_ (B.length str);;

let receive_of_channel ch n = buff_consume_data ch sock_feeder n;;

let close_in ch = Unix.close (get_descr ch.chan);;

let close_out ch = Unix.close (get_descr ch);;

let to_channel ch x _ =
  let s = Marshal.to_string x [] in
  send_bytes ch (bos s)
;;

let of_channel ch =
  try
    let hs = nat_marshal_header_size in
    let header = receive_of_channel ch hs in
    let sz = nat_marshal_data_size header _0_ in
    let data = receive_of_channel ch sz in
    nat_marshal_from_string (bts header ^ bts data) _0_
  with
  | Failure msg -> ioerr ("IO.of_channel: " ^ msg)
;;

let in_channel_of_ic ch = Unix.in_channel_of_descr (get_descr ch.chan);;
let out_channel_of_oc ch = Unix.out_channel_of_descr (get_descr ch);;

let ic_of_in_channel ic =
  let fd = Unix.descr_of_in_channel ic in
  bufferize (Fd fd)
;;

let oc_of_out_channel oc =
  let fd = Unix.descr_of_out_channel oc in
  Fd fd
;;

let channels_of_sock sock =
  let socket = Socket { fd = sock; addr = Unix.getsockname sock; } in
  (bufferize socket, socket)
;;

let stdin = bufferize (Fd Unix.stdin);;
let stdout = Fd Unix.stdout;;

(* Unix sockets helpers *)

let create_socket () = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0;;

let get_port = function
  | Unix.ADDR_INET (_, port) ->
    assert (port >= 0);
    (* Now, we have port >= 0.
       Therefore, the next Nat.unsafe_of_int will be safe. *)
    Nat.unsafe_of_int port
  | Unix.ADDR_UNIX _ -> raise Not_found
;;

let get_addr = function
  | Unix.ADDR_INET (a, _) -> a
  | Unix.ADDR_UNIX _ -> raise Not_found
;;

let connect addr =
  let fd = create_socket () in
  try
    Unix.connect fd addr;
    channels_of_sock fd
  with
  | Unix.Unix_error (e, _, _) ->
    Unix.close fd;
    ioerr ("IO.connect: " ^ Unix.error_message e)
;;

let connect_fd addr =
  let fd = create_socket () in
  begin
    try Unix.connect fd addr with
    | Unix.Unix_error (e, _, _) ->
      Unix.close fd;
      ioerr ("IO.connect_fd: " ^ Unix.error_message e)
  end;
  fd
;;

let create_bound_socket port =
  let port =
    match port with
    | None -> _0_
    | Some p -> p in
  let sock = create_socket () in
  try
    Unix.bind sock (Unix.ADDR_INET (Unix.inet_addr_any, Nat.to_int port));
    Unix.listen sock 1;
    (sock, Unix.getsockname sock)
  with
  | Unix.Unix_error (e, _, _) ->
    Unix.close sock;
    ioerr ("IO.create_bound_socket: " ^ Unix.error_message e)
;;

let wait_connection sock closeb =
  let (fd, addr) = Unix.accept sock in
  if closeb then Unix.close sock;
  (channels_of_sock fd, get_addr addr)
;;

(* Process management *)

type perms = Nat.t;;

type fork_data =
   | Child
   | Parent of pid
;;

let fork () =
  flush stderr; (* Avoid duplicating output *)
  try
    let pid = nat_unix_fork () in
    if pid = _0_ then Child else
    Parent pid
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.fork: " ^ Unix.error_message e)
;;

let open_process_in s =
  try ic_of_in_channel (Unix.open_process_in s) with
  | Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_process_in: " ^ Unix.error_message e)
;;

let open_process s =
  try
    let (ic, oc) = Unix.open_process s in
    (ic_of_in_channel ic, oc_of_out_channel oc)
  with
  | Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_process: " ^ Unix.error_message e)
;;

(* Files *)

let open_in_out str =
  let fd =
    try
      intr_loop
        (fun () -> Unix.openfile str [ Unix.O_RDWR; Unix.O_SYNC; ] 0)
    with
    | Unix.Unix_error (e, _, _) ->
      ioerr ("IO.open_in_out: " ^ Unix.error_message e) in
  (bufferize (Fd fd), Fd fd)
;;

let open_in str =
  let fd =
    try
      intr_loop
        (fun () -> Unix.openfile str [ Unix.O_RDONLY; Unix.O_SYNC; ] 0)
    with
    | Unix.Unix_error (e, _, _) ->
      ioerr ("IO.open_in: " ^ Unix.error_message e) in
  bufferize (Fd fd)
;;

let open_default_perms = Nat.unsafe_of_int 0o640;;

let open_out str =
  let fd =
    try
      intr_loop
        (fun () ->
          nat_unix_openfile
            str [ Unix.O_WRONLY; Unix.O_SYNC; ] open_default_perms)
    with
    | Unix.Unix_error (e, _, _) ->
      ioerr ("IO.open_out: " ^ Unix.error_message e) in
  Fd fd
;;

(* Fifo and pipes *)

let pipe () =
  let (ic, oc) =
    try Unix.pipe () with
    | Unix.Unix_error (e, _, _) ->
      ioerr ("IO.pipe: " ^ Unix.error_message e) in
  (bufferize (Fd ic), Fd oc)
;;

let fifo_default_perms = Nat.unsafe_of_int 0o640;;

let mk_fifo k str =
  let rec loop f =
    try f () with
    | Unix.Unix_error (Unix.EINTR, _, _) -> loop f
    | Unix.Unix_error (Unix.EEXIST, _, _) ->
      (* Fifos are named after pids, thus an already existing fifo is
       necessarily some remaining garbage, so remove it. *)
      Unix.unlink str;
      loop f in
  begin
    try loop (fun () -> nat_unix_mkfifo str fifo_default_perms) with
    | Unix.Unix_error (e, _, _) ->
      ioerr ("IO.mk_fifo: " ^ Unix.error_message e)
  end;
  k str
;;

let openfifo_in str = mk_fifo open_in str;;

let openfifo_out str = mk_fifo open_out str;;
