(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.28 2017/01/04 11:58:29 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module A = Uarray;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module Pic = Upph_id_card;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let nat_uaref_make_aref mk_aref_int name init =
  let { Ar.get; set; } = mk_aref_int name (Nat.to_int init) in
  (* We will always have get () >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  let get () = Nat.unsafe_of_int (get ()) in
  let set n = set (Nat.to_int n) in
  { Ar.get; set; }
;;

let qual = Perv.qualify "Options";;
let invalid_arg = Perv.invalid_arg_qual qual;;

let nat_set set n =
  if n < 0 then invalid_arg "nat_set" "negative number" else
  (* Now, we have n >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  set (Nat.unsafe_of_int n)
;;

module Opts = Tools_options;;

open Node_types;;

(* Command-line options. *)

let mode_of_string = function
  | "master" -> Master_mode
  | "slave" -> Slave_mode
  | s -> failwith (Printf.sprintf "Unbound mode %S" s)
;;

let string_of_mode = function
  | Master_mode -> "master"
  | Slave_mode -> "slave"
;;

let mode_name = "mode"
and slave_number_name = "slave number"
and slave_port_name = "slave port"
and sklml_port_name = "Sklml port"
and local_address_name = "local address"
and master_address_name = "master address"
and sklml_program_path_name = "Sklml program path"
and color_name = "color"
and dynamic_link_name = "dynamic link"
;;

let init_mode = Master_mode
and init_slave_number = _1_
and default_slave_port = _0_
and default_sklml_port = _0_
and default_local_address = Unix.inet_addr_loopback
;;
let init_master_address = Unix.ADDR_INET (default_local_address, 0)
and init_sklml_program_path = ""
and init_color = 1
and default_dynamic_link = true
;;

let (get_mode, set_mode) =
  let { Ar.get; set; } = Ar.make_aref_unique mode_name init_mode in
  (get, Perv.compose set mode_of_string)
;;

let incompatible_option_msg mode option_name =
  let s_mode = string_of_mode mode in
  Printf.sprintf "option %S is incompatible with %S mode" option_name s_mode
;;

let check_master_mode option_name =
  match get_mode () with
  | Master_mode -> ()
  | Slave_mode as mode ->
    let msg = incompatible_option_msg mode option_name in
    raise (Arg.Bad msg)

and check_slave_mode option_name =
  match get_mode () with
  | Master_mode as mode ->
    let msg = incompatible_option_msg mode option_name in
    raise (Arg.Bad msg)
  | Slave_mode -> ()
;;

let gen_mk_aref check_mode mk_aref name default =
  let { Ar.get; set; } = mk_aref name default in
  (get, (fun arg -> check_mode name; set arg))
;;

let mk_master_aref mk_aref = gen_mk_aref check_master_mode mk_aref
and mk_slave_aref mk_aref = gen_mk_aref check_slave_mode mk_aref
;;

let (get_slave_number, set_slave_number) =
  mk_master_aref
    (nat_uaref_make_aref Ar.make_aref_unique_positive_int)
    slave_number_name
    init_slave_number

and (get_slave_port, set_slave_port) =
  mk_master_aref
    (nat_uaref_make_aref Ar.make_aref_single_nonnegative_int)
    slave_port_name
    default_slave_port

and (get_sklml_port, set_sklml_port) =
  mk_master_aref
    (nat_uaref_make_aref Ar.make_aref_single_nonnegative_int)
    sklml_port_name
    default_sklml_port

and (get_local_address, set_local_address) =
  let { Ar.get; set; } =
    Ar.make_aref_single local_address_name default_local_address in
  (get, Perv.compose set Unix.inet_addr_of_string)

and (get_master_address, set_master_address) =
  let (get, set) =
    mk_slave_aref
      Ar.make_aref_unique
      master_address_name
      init_master_address in
  (get, Perv.compose set IO.read_sockaddr)

and (get_sklml_program_path, set_sklml_program_path) =
  mk_slave_aref
    Ar.make_aref_unique
    sklml_program_path_name
    init_sklml_program_path

and (get_color, set_color) =
  mk_slave_aref
    Ar.make_aref_unique_positive_int
    color_name
    init_color

and (get_dynamic_link, set_dynamic_link) =
  mk_slave_aref
    Ar.make_aref_single
    dynamic_link_name
    default_dynamic_link
;;

let get_master_node () =
  let local_addr = get_local_address ()
  and nslaves = get_slave_number ()
  and slave_port = get_slave_port ()
  and sklml_port = get_sklml_port () in
  {
   mn_local_address = local_addr;
   mn_slave_number = nslaves;
   mn_slave_port = slave_port;
   mn_sklml_port = sklml_port;
  }

and get_slave_node () =
  let local_addr = get_local_address ()
  and master_addr = get_master_address ()
  and path = get_sklml_program_path ()
  and color = get_color ()
  and dynlink = get_dynamic_link () in
  {
   sn_local_address = local_addr;
   sn_master_address = master_addr;
   sn_sklml_program_path = path;
   sn_color = color;
   sn_dynamic_link = dynlink;
  }
;;

let get_node () =
  match get_mode () with
  | Master_mode -> Master_node (get_master_node ())
  | Slave_mode -> Slave_node (get_slave_node ())
;;

let mode_lok = Ok.long_option_key_of_string mode_name
and slave_number_lok = Ok.long_option_key_of_string slave_number_name
and slave_port_lok = Ok.long_option_key_of_string slave_port_name
and sklml_port_lok = Ok.long_option_key_of_string sklml_port_name
and local_address_lok = Ok.long_option_key_of_string local_address_name
and master_address_lok = Ok.long_option_key_of_string master_address_name
and sklml_program_path_lok =
  Ok.long_option_key_of_string sklml_program_path_name
and color_lok = Ok.long_option_key_of_string color_name
and dynamic_link_lok = Ok.long_option_key_of_string dynamic_link_name
;;

let mode_rco =
  {
   Co.key = mode_lok;
   spec = Arg.String set_mode;
   doc =
     Printf.sprintf
       "master|slave\n\
        \tSelect either master or slave mode (mandatory).";
  }

and slave_number_rco =
  {
   Co.key = slave_number_lok;
   spec = Arg.Int (nat_set set_slave_number);
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet number of slave nodes to <nat> (M, mandatory).";
  }

and slave_port_rco =
  {
   Co.key = slave_port_lok;
   spec = Arg.Int (nat_set set_slave_port);
   doc =
     Printf.sprintf
       "<port>\n\
        \tSet Unix port number to connect slave nodes to <port> (M).\n\
        \tDefault value is %d."
       (Nat.to_int default_slave_port);
  }

and sklml_port_rco =
  {
   Co.key = sklml_port_lok;
   spec = Arg.Int (nat_set set_sklml_port);
   doc =
     Printf.sprintf
       "<port>\n\
        \tSet Unix port number to connect Sklml programs to <port> (M).\n\
        \tDefault value is %d"
       (Nat.to_int default_sklml_port);
  }

and local_address_rco =
  {
   Co.key = local_address_lok;
   spec = Arg.String set_local_address;
   doc =
     Printf.sprintf
       "<addr>\n\
        \tSet local address to <addr> (M|S).\n\
        \tDefault value is %S"
       (Unix.string_of_inet_addr default_local_address);
  }

and master_address_rco =
  {
   Co.key = master_address_lok;
   spec = Arg.String set_master_address;
   doc =
     Printf.sprintf
       "<addr>\n\
        \tSet master address to which connect current slave node to <addr>\n\
        \t(S, mandatory).";
  }

and sklml_program_path_rco =
  {
   Co.key = sklml_program_path_lok;
   spec = Arg.String set_sklml_program_path;
   doc =
     Printf.sprintf
       "<path>\n\
        \tSet path of Sklml program to run on current slave node to <path>\n\
        \t(S, mandatory).";
  }

and color_rco =
  {
   Co.key = color_lok;
   spec = Arg.Int set_color;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet color of current slave node to <nat> (S, mandatory).";
  }

and dynamic_link_rco =
  {
   Co.key = dynamic_link_lok;
   spec = Arg.Bool set_dynamic_link;
   doc =
     Printf.sprintf
       "<bool>\n\
        \tEnable/disable dynamic link with OCaml libraries listed in\n\
        \tSKLML_DLINK environment variable (S).\n\
        \tDefault value is %B"
       default_dynamic_link;
  }
;;

let command_kind = "parallel launcher daemon"
and command_suffix = "pard"
;;

let rcos =
  [
   mode_rco;
   slave_number_rco;
   slave_port_rco;
   sklml_port_rco;
   local_address_rco;
   master_address_rco;
   sklml_program_path_rco;
   color_rco;
   dynamic_link_rco;
  ]
;;

let project_name = Sklml_version.pph_id_card.Pic.project_name;;

let sklml_command_kind = Printf.sprintf "%s %s" project_name command_kind

let command_title = Printf.sprintf "%s%s" project_name command_suffix;;

let phase_mode = Co.Phase mode_name;;

let add_options () =
  Co.application_add_all_rco rcos;
  Co.phase_add phase_mode mode_lok;
;;

add_options ();;

Pic.add_app_version_options Sklml_version.pph_id_card None;;

(* Sys.argv is never empty.
   Therefore, the next A.unsafe_get will be safe. *)
let cmd_name = A.unsafe_get Sys.argv _0_;;
Co.set_usage_msg
  (Format.sprintf
     "@[<v>\
      @ \
      %s -- %s@ \
      @ \
      Usage:@ \
      @ \
      %s <arguments>...@ \
      @ \
      Argument \"-mode\" is mandatory, it is parsed first.@ \
      Other arguments are tagged (M) for master mode and (S) for slave mode.@ \
      Some of them are mandatory.@ \
      @ \
      Arguments are:\
      @]@."
     cmd_name sklml_command_kind
     cmd_name)
;;

(* Report/debug flags and warning/error messages. *)

let services = Opts.add_tagged_services command_title;;

let format_sprint print =
  print Format.str_formatter;
  Format.flush_str_formatter ()
;;

let print_ident_value ident printer ppf value =
  Format.fprintf ppf "@[<v 2>@ @[<v 2>%s =@ %a@]@]@." ident printer value
;;

let sprint_ident_value ident printer value =
  format_sprint (fun ppf -> print_ident_value ident printer ppf value)
;;

(* Parsing the command line. *)

(* We need here a specific treatment of excpetions, hence the use of
 parse*_argv functions inside a try-with construction in function Main.main. *)
let parse () =
    let argv = Sys.argv in
    let argv = Co.parse_all_services_argv argv in
    let argv = Co.parse_phase_argv phase_mode argv in
    Co.parse_argv argv;
    get_node ()
;;
