(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parfun.ml,v 1.73 2017/01/04 11:58:29 fclement Exp $ *)

module Nat = Unatural;;
module A = Uarray;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let nat_unix_getpid () =
  let pid = Unix.getpid () in
  assert (pid >= 0);
  (* Now, we have pid >= 0.
     Therefore, the next Nat.unsafe_of_int will be safe. *)
  Nat.unsafe_of_int pid
;;

exception Parfun_error;;

type ('x, 'a, 'b) worker =
  {
   io : IO.t;
   f : 'a Sklstream.t -> 'b Sklstream.t;
  }
;;

type master = { daemon_io : IO.t; };;

type ('x, 'a, 'b) startup_ret =
  | Master of master
  | Worker of ('x, 'a, 'b) worker
;;

type t =
  { parfun : 'a 'b. ('a, 'b) Skl.t -> 'a Sklstream.t -> 'b Sklstream.t; }
and parfun = t
;;

let request_daemon io req =
  match Skl_comm.request_daemon io req with
  | None
  | Some Skl_comm.Failure -> raise Parfun_error
  | Some (Skl_comm.Success addr) -> addr
;;

(* Core of the parallel implementation *)

let rec start_tree : 'a 'b. IO.t -> ('a, 'b) Skl.t -> Iotree.ioatree =
  fun io -> function

    | Sk.Skl.Atom atom ->
      let f_atom params llexp color =
        let spawn_request =
          let (p, id) = Expor.get_exportable (params, llexp) in
          Skl_comm.Atom_req (p, id, color) in
        let addr = request_daemon io spawn_request in
        Iotree.IOALeaf addr in
      Sk.with_atom atom { Sk.f_atom; }

    | Sk.Skl.Id id_skl ->
      let f_id _ = Iotree.IOALeafId in
      Sk.with_id id_skl { Sk.f_id; }

    | Sk.Skl.Loop loop ->
      let f_loop _ ptree tree =
        let addr_tree1 = start_tree io ptree in
        let addr_tree2 = start_tree io tree in
        let spawn_request = Skl_comm.Loop_req (addr_tree1, addr_tree2) in
        let addr = request_daemon io spawn_request in
        Iotree.IOALeaf addr in
      Sk.with_loop loop { Sk.f_loop; }

    | Sk.Skl.Par par ->
      let f_par t1 t2 =
        let addr_tree1 = start_tree io t1 in
        let addr_tree2 = start_tree io t2 in
        Iotree.IOABranch (addr_tree1, addr_tree2) in
      Sk.with_par par { Sk.f_par; }

    | Sk.Skl.Pipe pipe ->
      let f_pipe t1 t2 =
        let addr_tree1 = start_tree io t1 in
        let addr_tree2 = start_tree io t2 in
        let spawn_request = Skl_comm.Pipe_req (addr_tree1, addr_tree2) in
        let addr = request_daemon io spawn_request in
        Iotree.IOALeaf addr in
      Sk.with_pipe pipe { Sk.f_pipe; }

    | Sk.Skl.Prod prod ->
      let f_prod _ _ skl1 skl2 =
        let addr_tree1 = start_tree io skl1 in
        let addr_tree2 = start_tree io skl2 in
        Iotree.IOASplit (addr_tree1, addr_tree2) in
      Sk.with_prod prod { Sk.f_prod; }

    | Sk.Skl.Rails rails ->
      let f_rails ska =
        let addr_treev = A.map (start_tree io) ska in
        Iotree.IOALeafs addr_treev in
      Sk.with_rails rails { Sk.f_rails; }

    | Sk.Skl.Sum sum ->
      let f_sum _ t1 t2 =
        let addr_tree1 = start_tree io t1 in
        let addr_tree2 = start_tree io t2 in
        Iotree.IOASumSplit (addr_tree1, addr_tree2) in
      Sk.with_sum sum { Sk.f_sum; }
;;

let parfun io skl is =
  let tree =
    Iotree.iotree_of_ioatree (* Connects the tree *)
      (start_tree io skl) in
  match IO.fork () with
  | IO.Parent _ -> Iotree.receive_in_tree tree
  | IO.Child ->
    Iotree.send_in_tree tree is;
    exit 0
;;

type arg_master = { addr : Unix.sockaddr; }
and arg_worker = { f_id : Nat.t; }
;;

type parsed_args =
  | Arg_master of arg_master
  | Arg_worker of arg_worker
;;

let (get_mode, set_master_mode, set_slave_mode) =
  let { Ar.get = get_mode; set = set_mode; } =
    Ar.make_aref_mandatory
      "Sklml parallel mode"
      (Arg_worker { f_id = _0_; }) in
  let set_master_mode x =
    set_mode (Arg_master { addr = IO.read_sockaddr x; }) in
  let set_slave_mode x = set_mode (Arg_worker { f_id = Nat.of_int x; }) in
  (get_mode, set_master_mode, set_slave_mode)
;;

let phase_sklml_par = Co.Phase "Sklml parallel";;

let master_ok = Ok.of_string "-master"
and atom_ok = Ok.of_string "-atom"
;;

let master_rco =
  {
   Co.key = master_ok;
   spec = Arg.String set_master_mode;
   doc = "";
  }

and atom_rco =
  {
   Co.key = atom_ok;
   spec = Arg.Int set_slave_mode;
   doc = "";
  }
;;

let rcos =
  [
   master_rco;
   atom_rco;
  ]
;;

let add_options () =
  Co.library_add_all_rco rcos;
  Co.phase_add_all phase_sklml_par [ master_ok; atom_ok; ];
;;

let fail m = Log.say Log.Err "%s" m; exit 2;;

let parse () =
  try Co.parse_phase phase_sklml_par with
  | Failure m -> fail m
;;

let startup () =
  add_options ();
  parse ();
  let mode =
    try get_mode () with
    | Failure m -> fail m in
  match mode with
  | Arg_master { addr = addr; } ->
    let io = IO.connect addr in
    Master { daemon_io = io; }
  | Arg_worker { f_id = f_id; } ->
    Co.parse ();
    let f_gen = Expor.unsafe_retreive f_id in
    let (fifo_in, fifo_out) =
      Skl_comm.open_skl_fifo false (nat_unix_getpid ()) in
    let params = IO.of_channel fifo_in in
    let f = f_gen params in
    let (sock, addr) = IO.create_bound_socket None in
    let port = IO.get_port addr in
    Unix.listen sock 1; (* listen before giving the port *)
    IO.to_channel fifo_out port true;
    let io, _ = IO.wait_connection sock true in
    Worker { io = io; f = f; }
;;
(* Startup function, if -atom N is spotted in [Sys.argv] then [startup ()]
 runs the atom code, otherwise, starts as master and connects to the Sklml
 parallel launcher daemon given as argument. *)

let run_worker w =
  let { io = (ic, oc); f = f; } = w in
  Sklstream.to_io_channel (f (Sklstream.of_io_channel ic)) oc
;;

let pardo f =
  match startup () with
  | Master { daemon_io = io; } -> f { parfun = fun x -> parfun io x; }
  | Worker w ->
    run_worker w;
    Log.say (Log.Log _1_) "Worker job done."
;;
(* Entry point. *)
