(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parser_plus.ml,v 1.19 2017/05/16 18:45:04 fclement Exp $ *)

(* More parsing operations. *)

module Nat = Unatural;;
module L = Ulist;;
module S = Ustring;;

let _0_ = Nat._0_
and _10_ = Nat.unsafe_of_int 10
and _45_ = Nat.unsafe_of_int 45
and _48_ = Nat.unsafe_of_int 48
and _57_ = Nat.unsafe_of_int 57
and _65_ = Nat.unsafe_of_int 65
and _90_ = Nat.unsafe_of_int 90
and _95_ = Nat.unsafe_of_int 95
and _97_ = Nat.unsafe_of_int 97
and _122_ = Nat.unsafe_of_int 122
and _255_ = Nat.unsafe_of_int 255
;;

let rec nat_range min max =
  if min = max then [ min; ] else
  min :: (nat_range (Nat.succ min) max)
;;

open Parser;;

let char_minus = _45_
and char_0 = _48_
and char_9 = _57_
and char_A = _65_
and char_Z = _90_
and char_underscore = _95_
and char_a = _97_
and char_z = _122_
;;

(* alphas is [a-zA-Z_-] *)
let nat_alphas =
  nat_range char_a char_z @
  nat_range char_A char_Z @
  [ char_underscore; char_minus; ]

(* nums is [0-9] *)
and nat_nums = nat_range char_0 char_9

and nat_all_chars = nat_range _0_ _255_
;;

let alphas = L.map Nat.to_char nat_alphas
and nums = L.map Nat.to_char nat_nums
and all_chars = L.map Nat.to_char nat_all_chars
;;

let spaces = [ '\r'; '\n'; '\t'; ' '; ];;

(* Helper functions. *)

let p_ignore p = p >> return ();;

let notc l = L.filter (fun c -> not (L.mem c l)) all_chars;;

let to_num l =
  let rec loop = function
    | [] -> 0
    | h :: t -> Char.code h - Char.code '0' + loop t * 10 in
  loop (L.rev l)
;;

let to_nat l =
  let digit_shift = Nat.of_char '0' in
  let rec loop = function
    | [] -> _0_
    | h :: t ->
      let h = Nat.of_char h in
      assert (digit_shift <= h);
      (* Now, we have digit_shift <= h.
         Therefore, the following Nat.sub will not fail. *)
      let d = Nat.sub h digit_shift in
      (* Char.code h - Char.code '0' + loop t * 10. *)
      Nat.add d (Nat.mul (loop t) _10_) in
  loop (L.rev l)
;;

(* Common parsers. *)

let p_letter = any (L.map char alphas);;
let p_digit = any (L.map char nums);;
let p_space = any (L.map char spaces);;

let p_ident = many1 p_letter >>= fun l -> return (S.gather l);;

let p_integer = many1 p_digit >>= fun l -> return (to_num l)
and p_natural = many1 p_digit >>= fun l -> return (to_nat l)
;;

let p_word s =
  let len = S.length s in
  let rec loop i =
    if i >= len then return () else
    (* Now, we have 0 <= i < len.
       Therefore, the next S.unsafe_get will be safe. *)
    char (S.unsafe_get s i) >> loop (Nat.succ i) in
  loop _0_
;;
